@@ -2,16 +2,15 @@
 
 import asyncio
 import weakref
-from functools import partial
-from random import randint
-from typing import List, Union, Tuple, Iterable, Set
-
 from aiokafka import AIOKafkaConsumer, ConsumerRebalanceListener, ConsumerRecord
+from functools import partial
 from humanize import naturalsize
 from kafka import TopicPartition
 from patchwork.core import AsyncSubscriber, Task
 from patchwork.core.utils import AsyncQueue
 from pydantic import root_validator
+from random import randint
+from typing import List, Union, Tuple, Iterable, Set
 
 from .common import SECOND, get_node_id, MINUTE
 
@@ -96,9 +95,11 @@ class AsyncKafkaSubscriber(AsyncSubscriber):
         self._freezer = None
 
         self._getters_queue = AsyncQueue(maxsize=1)
-        self._task_meta = {}
         self._subscribed = asyncio.Event()
 
+        self._in_process = set()
+        self._resumed = asyncio.Event()
+
     def __repr__(self):
         res = super().__repr__()
         return f"<{res[1:-1]} [brokers={','.join(self.settings.kafka_hosts)}]>"
@@ -113,9 +114,6 @@ class AsyncKafkaSubscriber(AsyncSubscriber):
     def session_timeout(self):
         return self.settings.health_check_ms*MAX_TIME_MULTIPLIER
 
-    def _on_task_ref_removed(self, ref, uuid):
-        self._task_meta.pop(uuid)
-
     async def _start(self):
         """
         Starts the consumer and subscribes for given topics
@@ -171,24 +169,14 @@ class AsyncKafkaSubscriber(AsyncSubscriber):
             **dict(meta.headers),
             'queue_name': meta.topic,
         })
-        # keep some metadata about task for commit
-        self._task_meta[task.uuid] = (
-            meta.topic,
-            meta.partition,
-            meta.offset,
-            weakref.ref(task, partial(self._on_task_ref_removed, uuid=task.uuid))
-        )
+
+        task._kafka_subscriber = (meta.topic, meta.partition, meta.offset)
         return task
 
-    def _get_message_meta(self, task: Task):
-        meta = self._task_meta.pop(task.uuid, None)
-        if not meta:
-            self.logger.error(f"no metadata for given task '{task.task_type} (uuid={task.uuid})'")
-            raise KeyError()
-
-        tp = TopicPartition(meta[0], meta[1])
-        offset = meta[2]
-        return tp, offset
+    def _resume_partition(self, tp: TopicPartition):
+        self._in_process.discard(tp)
+        self._resumed.set()
+        self.logger.debug(f"{tp} resumed")
 
     async def commit(self, task: Task, timeout: float = None):
         """
@@ -198,7 +186,9 @@ class AsyncKafkaSubscriber(AsyncSubscriber):
         :param timeout:
         :return:
         """
-        tp, offset = self._get_message_meta(task)
+        assert hasattr(task, '_kafka_subscriber'), "cannot commit task not received by this subscriber"
+        tp = TopicPartition(task._kafka_subscriber[0], task._kafka_subscriber[1])
+        offset = task._kafka_subscriber[2]
 
         try:
             # commit message
@@ -213,16 +203,17 @@ class AsyncKafkaSubscriber(AsyncSubscriber):
             return False
         finally:
             # resume partition
-            self.consumer.resume(tp)
-            self.logger.debug(f"{tp} resumed")
+            self._resume_partition(tp)
 
     async def rollback(self, task: Task, timeout: float = None):
-        tp, offset = self._get_message_meta(task)
+        assert hasattr(task, '_kafka_subscriber'), "cannot commit task not received by this subscriber"
+        tp = TopicPartition(task._kafka_subscriber[0], task._kafka_subscriber[1])
+        offset = task._kafka_subscriber[2]
+
         # seek just before rollback message
         self.consumer.seek(tp, offset)
         # and resume partition
-        self.consumer.resume(tp)
-        self.logger.debug(f"{tp} resumed")
+        self._resume_partition(tp)
 
     async def handle_partitions_revoke(self, revoked):
         """
@@ -294,12 +285,30 @@ class AsyncKafkaSubscriber(AsyncSubscriber):
         Without periodic pooling consumer will be marked as dead by coordinator.
         :return:
         """
+        self._in_process.clear()
         while not self.is_stopping:
 
-            msg: ConsumerRecord = await self.consumer.getone()
+            # prioritize TP, not in process firstly
+            partitions = sorted(self.consumer.assignment(), key=lambda tp: tp in self._in_process)
+            msg: ConsumerRecord = await self.consumer.getone(*partitions)
+            tp = TopicPartition(topic=msg.topic, partition=msg.partition)
+
+            if tp in self._in_process:
+                # this TP has already a task in process, so there is no tasks to fetch
+                # on partitions which are not in process (were prioritized above)
+                self.consumer.seek(tp, msg.offset)
+                try:
+                    # wait until something resumed or short period, so if something arrives at
+                    # not in process TP it will be processed
+                    await asyncio.wait_for(self._resumed.wait(), timeout=0.250)
+                except TimeoutError:
+                    pass
+                else:
+                    self._resumed.clear()
+                continue
 
             try:
-                # wait for getter not longer then half pool interval
+                # wait for getter no longer then half pool interval
                 getter: asyncio.Future = await self._getters_queue.get(
                     timeout=self.settings.poll_interval_ms / 2 / SECOND
                 )
@@ -307,24 +316,20 @@ class AsyncKafkaSubscriber(AsyncSubscriber):
                 # seek back to get same message again
                 self.consumer.seek(TopicPartition(topic=msg.topic, partition=msg.partition), msg.offset)
                 continue
-            else:
-                self._getters_queue.task_done()
-                if getter.cancelled():
-                    # seek back to get same message again
-                    self.consumer.seek(TopicPartition(topic=msg.topic, partition=msg.partition), msg.offset)
-                    continue
 
-                getter.set_result(msg)
+            self._getters_queue.task_done()
+            if getter.cancelled():
+                # seek back to get same message again
+                self.consumer.seek(TopicPartition(topic=msg.topic, partition=msg.partition), msg.offset)
+                continue
+
+            getter.set_result(msg)
+            self._in_process.add(tp)
+            self.logger.debug(f"{tp} paused")
 
             msg_key = f'{msg.topic}:{msg.partition}@{msg.offset}'
             self.logger.debug(f"Message '{msg_key}' received ({naturalsize(len(msg.value))})")
 
-            # pause the topic-partition to make sure that no more messages will be fetched
-            # from this topic-partition
-            tp = TopicPartition(topic=msg.topic, partition=msg.partition)
-            self.consumer.pause(tp)
-            self.logger.debug(f"{tp} paused")
-
     def _fetch_messages_done(self, fut):
         self._fetcher = None
         if fut.cancelled():
@@ -443,6 +448,4 @@ class AssignKafkaSubscriber(AsyncKafkaSubscriber):
         else:
             # no commits if no consumer group, just resume paused TP
             tp, _ = self._get_message_meta(task)
-
-            self.consumer.resume(tp)
-            self.logger.debug(f"{tp} resumed")
+            self._resume_partition(tp)
